/*
 * This file is part of the openKropki project.
 *
 * Copyright (C) 2014, 2015, Mateusz Viste
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>   /* printf(), puts() */
#include <stdlib.h>  /* rand(), srand() */
#include <string.h>  /* memset */
#include <time.h>    /* time() */
#include <sys/time.h> /* gettimeofday() */
#include <SDL2/SDL_thread.h>

#include "ai.h"  /* include self for control */


struct minimax_t {
  int x, y;
  int score;
};

struct threaddata_t {
  struct minimax_t *minimax;
  struct kropki_game *game;
  int ai_player;
};


static int distance(int x1, int y1, int x2, int y2) {
  int xdist, ydist;
  if (x1 > x2) {
    xdist = x1 - x2;
  } else {
    xdist = x2 - x1;
  }
  if (y1 > y2) {
    ydist = y1 - y2;
  } else {
    ydist = y2 - y1;
  }
  return(xdist + ydist);
}


/* there are some specific situations when a simple hard-wired logic can
 * detect things NOT to do - this function provides a VETO against such
 * well-known bad situations. */
static int is_veto_situation(int ai_player, struct kropki_game *game, int x, int y) {
  int i, t, a, b, c, d;
  /* do not put your soldier between three oponents */
  i = 0; /* i will count the number of opponents around */
  t = 0; /* t is the number of my soldiers */
  a = kropki_getpos(game, x, y - 1) & 3;
  b = kropki_getpos(game, x, y + 1) & 3;
  c = kropki_getpos(game, x - 1, y) & 3;
  d = kropki_getpos(game, x + 1, y) & 3;
  if (a == 3 - ai_player) i++;
  if (b == 3 - ai_player) i++;
  if (c == 3 - ai_player) i++;
  if (d == 3 - ai_player) i++;
  if (a == ai_player) t++;
  if (b == ai_player) t++;
  if (c == ai_player) t++;
  if (d == ai_player) t++;
  if ((i > 2) && (t == 0)) {
    printf("%d:%d is a VETO!\n", x, y);
    return(1);
  }
  return(0);
}


static void ai_idiot(int ai_player, struct kropki_game *game) {
  int x, y, attempts, found = 0;
  for (attempts = 0;; attempts++) {
    x = rand() % 5;
    y = rand() % 5;
    x = game->lastmovex - 2 + x;
    y = game->lastmovey - 2 + y;
    if ((x < 0) || (y < 0)) continue;
    if (x >= game->field_width) x = game->field_width - 1;
    if (y >= game->field_height) y = game->field_height - 1;
    if (kropki_getpos(game, x, y) == kropki_EMPTY) {
      found = 1;
      break;
    }
    if (attempts >= 100) { /* after 100 tries, stop */
      break;
    }
  }
  if (found == 0) {
    /* if unable to find anything randomnly, scan the field sequentially */
    if (kropki_findfirstfreefield(game, &x, &y) == 0) found = 1;
  }
  /* if a possible play move is found, do it */
  if (found != 0) {
      /* printf("AI PLAYS [%d,%d]\n", x, y); */
      kropki_setpos(game, x, y, ai_player);
    } else {
      puts("NOTHING FOUND!!!!!");
  }
}


static int ai_smart_minimax(int ai_player, struct kropki_game *game, struct minimax_t *minmax, int recur) {
  int bestscore = 0, bestmovex = -1, bestmovey = -1;
  int x, y, points[2], deltascore = 0, orgdelta;
  struct kropki_game *simgame;
  kropki_getpoints(game, points);
  orgdelta = points[ai_player - 1] - points[(3 - ai_player) - 1];
  for (y = 0; y < game->field_height; y++) {
    for (x = 0; x < game->field_width; x++) {
      int vetoflag = 0;
      /* if field not empty -> skip it */
      if (kropki_getpos(game, x, y) != kropki_EMPTY) continue;
      /* for secondary recursion, analyze only nearest fields */
      if ((recur >= 1) && (distance(x, y, game->lastmovex, game->lastmovey) > 3)) continue;
      if ((recur >= 2) && (distance(x, y, game->lastmovex, game->lastmovey) > 1)) continue;
      /* do the actual simulation */
      simgame = kropki_copygame(game);
      kropki_setposforce(simgame, x, y, ai_player);
      kropki_getpoints(simgame, points);
      deltascore = points[ai_player - 1] - points[(3 - ai_player) - 1];
      deltascore -= orgdelta;
      if (recur == 0) {
        deltascore *= 8; /* immediate consequences are much more important than those 2 or 3 moves away */
        vetoflag = is_veto_situation(3 - ai_player, game, x, y); /* check immediate veto positions */
      }
      if (recur < 2) {
        deltascore += ai_smart_minimax(ai_player, simgame, minmax, recur + 1);
      }
      if ((deltascore > bestscore) || (bestmovex < 0)) {
        /* printf("best move for ai %d found at [%d,%d] -> %d pts\n", ai_player, x, y, deltascore); */
        bestmovex = x;
        bestmovey = y;
        bestscore = deltascore;
      } else if ((deltascore == bestscore) && (vetoflag == 0)) { /* if same score, see if closer, unless VETOed */
        if (distance(x, y, game->lastmovex, game->lastmovey) < distance(bestmovex, bestmovey, game->lastmovex, game->lastmovey)) {
          /* printf("[%d,%d] is closer to [%d,%d]\n", x, y, game->lastmovex, game->lastmovey); */
          bestmovex = x;
          bestmovey = y;
        }
      }
      kropki_freegame(simgame);
    }
  }
  minmax->x = bestmovex;
  minmax->y = bestmovey;
  minmax->score = bestscore;
  return(bestscore);
}


static int ai_smart_minimax_thread(void *param) {
  struct threaddata_t *p = param;
  ai_smart_minimax(p->ai_player, p->game, p->minimax, 0);
  return(0);
}


static void ai_smart(int ai_player, struct kropki_game *game) {
  SDL_Thread* threadid[2];
  struct minimax_t mini, maxi;
  struct threaddata_t threadparam[2];
  struct timeval starttime, stoptime;
  unsigned long mstime;
  int playsafe;

  memset(&mini, 0, sizeof(mini));
  memset(&maxi, 0, sizeof(maxi));

  gettimeofday(&starttime, NULL);

  /* prepare and run 1st thread */
  threadparam[0].minimax = &maxi;
  threadparam[0].game = game;
  threadparam[0].ai_player = ai_player;
  threadid[0] = SDL_CreateThread(ai_smart_minimax_thread, "max", &threadparam[0]);
  /* prepare and run 2nd thread */
  threadparam[1].minimax = &mini;
  threadparam[1].game = game;
  threadparam[1].ai_player = 3 - ai_player;
  threadid[1] = SDL_CreateThread(ai_smart_minimax_thread, "min", &threadparam[1]);
  /* wait for the parallell threads to terminate */
  SDL_WaitThread(threadid[0], NULL);
  SDL_WaitThread(threadid[1], NULL);

  gettimeofday(&stoptime, NULL);
  mstime = (stoptime.tv_sec - starttime.tv_sec) * 1000;
  mstime += (stoptime.tv_usec - starttime.tv_usec) / 1000;
  printf("max: [%d,%d] -> %d pts\nmin: [%d,%d] -> %d pts\ncomputation time: %lu ms\n\n", mini.x, mini.y, mini.score, maxi.x, maxi.y, maxi.score, mstime);

  /* decide whether to play defensively or offensively */
  if (maxi.score > mini.score) {
    playsafe = 0;
  } else { /* if I'm afraid to loose points, check the bold move anyway */
    struct kropki_game *simgame;
    struct minimax_t newmini;
    memset(&newmini, 0, sizeof(newmini));
    playsafe = 1;
    if (maxi.score > 0) {
      /* simulate the bold move */
      simgame = kropki_copygame(game);
      ai_smart_minimax(3 - ai_player, simgame, &newmini, 0);
      /* recheck min now - is it still bad? */
      if (newmini.score <= maxi.score) {
        printf("it appears that playing boldy might be rewarding!\n\n");
        playsafe = 0;
      }
      kropki_freegame(simgame);
    }
  }

  /* make the move */
  if (playsafe != 0) {
    if (kropki_setpos(game, mini.x, mini.y, ai_player) != 0) puts("WOOT, UNABLE TO PLAY!!! [mini]");
  } else {
    if (kropki_setpos(game, maxi.x, maxi.y, ai_player) != 0) puts("WOOT, UNABLE TO PLAY!!! [maxi]");
  }
}


void ai_play(enum ai_model model, int ai_player, struct kropki_game *game) {
  static int rand_init = 0;
  /* init the random generator, if not inited yet */
  if (rand_init == 0) {
    rand_init = 1;
    srand(time(NULL));
  }

  /* verify first that we can make any move at all */
  if (kropki_findfirstfreefield(game, NULL, NULL) != 0) return;

  /* execute the AI now */
  switch (model) {
    case AI_IDIOT:
      ai_idiot(ai_player, game);
      break;
    case AI_SMART:
      ai_smart(ai_player, game);
      break;
    default:
      puts("UNKNOWN AI MODEL CALLED");
      break;
  }
}
