/*
 * openKropki
 *
 * Copyright (C) 2012, 2013, 2014, 2015, Mateusz Viste
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdint.h>

#include "ai.h"
#include "drv_gra.h"
#include "ui.h"
#include "kropki_core.h"
#include "version.h"


static void redraw_screen(struct kropki_game *game, int width, int height, int *scrollvector, char *msg) {
  int x, y, scores[2];
  struct kropki_wallelement *wallsptr;
  ui_drawKartka(width, height, scrollvector);
  /* draw walls */
  for (wallsptr = game->walls; wallsptr != NULL; wallsptr = wallsptr->next) {
    graph_drawWall(wallsptr->x1, wallsptr->y1, wallsptr->x2, wallsptr->y2, wallsptr->player, scrollvector);
  }
  /* draw dots */
  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      int fieldval;
      fieldval = kropki_getpos(game, x, y);
      if ((fieldval & 3) != 0) {
        int activeflag = 0;
        if ((game->lastmovex == x) && (game->lastmovey == y)) activeflag = 1;
        graph_drawKropka(x, y, fieldval & 3, activeflag, scrollvector);
      }
    }
  }
  /* apply the fade-out border */
  /* ui_applyfaderect(25); */
  /* Draw scores */
  kropki_getpoints(game, scores);
  ui_drawscores(scores, msg);
  /* refresh the screen */
  gra_refresh();
}



#define PLAY_WIDTH  30
#define PLAY_HEIGHT 20
#define WIN_WIDTH  720
#define WIN_HEIGHT 500
#define COLORPLAYER1 0x3030FF
#define COLORPLAYER2 0xFF0000

static void centerscrollvector(int *scrollvector, int kratkasize) {
  int w, h;
  gra_getwinsize(&w, &h);
  if (w > PLAY_WIDTH) {
      scrollvector[0] = (w - (PLAY_WIDTH + 1) * kratkasize) / 2;
    } else {
      scrollvector[0] = ((PLAY_WIDTH + 1) * kratkasize - w) / 2;
  }
  if (h > PLAY_HEIGHT) {
      scrollvector[1] = (h - (PLAY_HEIGHT + 1) * kratkasize) / 2;
    } else {
      scrollvector[1] = ((PLAY_HEIGHT + 1) * kratkasize - h) / 2;
  }
}


int main(void) {
  enum ui_gamemode_t gamemode;
  struct kropki_game *game;
  int exitflag = 0;
  int scrollvector[2] = {0, 0};
  int kratkasize = 30;

  gra_init(WIN_WIDTH, WIN_HEIGHT, 0, "openKropki", NULL, 0);

  while (exitflag == 0) {
    int skipnexteventcheck = 0;
    int redraw = 1;

    gamemode = ui_selectgamemode();
    if (gamemode == GAMEMODE_QUIT) break;

    if (gamemode == GAMEMODE_NET) {
      gra_clear();
      ui_throwmsg("NOT  IMPLEMENTED  YET,  SORRY!");
      ui_delay(1500);
      continue;
    }

    if (ui_setparams(PLAY_WIDTH, PLAY_HEIGHT, kratkasize, 2, 0xA0A0A0, COLORPLAYER1, COLORPLAYER2, 0xFEFEFE, kratkasize/10, kratkasize/4) != 0) return(0);

    game = kropki_newgame(PLAY_WIDTH, PLAY_HEIGHT);
    centerscrollvector(scrollvector, kratkasize);

    while (exitflag == 0) {
      enum ui_events_t event;
      int x, y, z, playerinputx, playerinputy;
      playerinputx = -1;
      playerinputy = -1;

      /* redraw screen if needed */
      if (redraw != 0) {
        char msg[64] = {0};
        /* check whether the game ended */
        if (kropki_findfirstfreefield(game, NULL, NULL) != 0) {
          int points[2], winner;
          char *winnername[] = {"NOBODY", "YOU", "OPPONENT"};
          kropki_getpoints(game, points);
          if (points[0] > points[1]) {
            winner = 1;
          } else if (points[0] < points[1]) {
            winner = 2;
          } else {
            winner = 0;
          }
          sprintf(msg, "%s WINS!", winnername[winner]);
        } else {
          switch (game->lastplayer) {
            case 1:
              sprintf(msg, "Opponent thinks...");
              break;
            case 2:
              sprintf(msg, "Your turn...");
              break;
          }
        }
        redraw_screen(game, PLAY_WIDTH, PLAY_HEIGHT, scrollvector, msg);
        redraw = 0;
      }

      /* poll events */
      if (skipnexteventcheck != 0) {
          event = Event_NONE;
          skipnexteventcheck = 0;
        } else {
          event = ui_getevent(&x, &y, &z, 0);
      }
      if (event == Event_QUIT) {
          exitflag = 1;
          break;
        } else if (event == Event_CLICK) {
          int fieldx, fieldy;
          if (z == 1) {
              graph_mousemap(x, y, &fieldx, &fieldy, scrollvector);
              if ((fieldx >= 0) && (fieldy >= 0) && (z == 1)) {
                playerinputx = fieldx;
                playerinputy = fieldy;
              }
            } else if (z == 2) {
              centerscrollvector(scrollvector, kratkasize);
              redraw = 1;
          }
        } else if (event == Event_SWEEP) {
          /* printf("SWEEP WITH BUTTON %d BY VECTOR (%d,%d)\n", z, x, y); */
          scrollvector[0] -= x;
          scrollvector[1] -= y;
          redraw = 1;
        } else if (event == Event_SCROLL) {
          if ((y > 0) && (kratkasize < 100)) kratkasize += 1;
          if ((y < 0) && (kratkasize > 5)) kratkasize -= 1;
          redraw = 1;
          ui_setparams(PLAY_WIDTH, PLAY_HEIGHT, kratkasize, 2, 0xA0A0A0, COLORPLAYER1, COLORPLAYER2, 0xFEFEFE, kratkasize/10, kratkasize/4);
        } else if (event != Event_NONE) {
          /* puts("got an unknown event"); */
          redraw = 1;
      }
      /* check for movements */
      if (game->lastplayer == 1) { /* AI plays */
          ui_changecursor(CURSOR_WAIT);
          ai_play(AI_SMART, 2, game);
          ui_changecursor(CURSOR_NORMAL);
          redraw = 1;
        } else { /* human plays */
          if (playerinputx >= 0) {
            kropki_setpos(game, playerinputx, playerinputy, 1);
            redraw = 1;
            skipnexteventcheck = 1; /* skip the next check of events to let the AI play ASAP */
          }
      }
    }
    kropki_freegame(game);
  }

  /* clean up and quit */
  gra_close();

  return(0);
}
