/*
 * This file is part of the openKropki project.
 * It provides abstract drawing functions based on SDL and SDL_gfx.
 *
 * Copyright (C) 2014, 2015, Mateusz Viste
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <SDL2/SDL.h>
#include <SDL2/SDL_ttf.h>
#include <stdio.h>            /* sprintf() */
#include <stdlib.h>           /* abs() */
#include "drv_gra.h"
#include "data.h"             /* contains the font data */
#include "ui.h"               /* include self for control */
#include "version.h"

static int kratkaSize = 0;
static int kratkaThickness = 0;
static unsigned int kratkaColorR = 0, kratkaColorG = 0, kratkaColorB = 0, kratkaColorA = 0xff;
static unsigned int playercolorR[3] = {0, 0, 0};
static unsigned int playercolorG[3] = {0, 0, 0};
static unsigned int playercolorB[3] = {0, 0, 0};
static unsigned int playercolorA[3] = {0, 0, 0};
static unsigned int kartkaBackgroundColorR = 0,  kartkaBackgroundColorG = 0, kartkaBackgroundColorB = 0, kartkaBackgroundColorA = 0;
static int wallThickness = 0;
static int kropkaRadius = 0;
static int kropkaActiveBorder = 0;
static int plansza_width = 0, plansza_height = 0;


int ui_setparams(int x, int y, int tkratkaSize, int tkratkaThickness, int tkratkaKolor, int tplayer1color, int tplayer2color, int tkartkaBackgroundColor, int twallThickness, int tkropkaRadius) {
  /* Save all parameters */
  plansza_width = x;
  plansza_height = y;
  kratkaSize = tkratkaSize;
  kratkaThickness = tkratkaThickness;
  kratkaColorR = (tkratkaKolor >> 16) & 0xff;
  kratkaColorG = (tkratkaKolor >> 8) & 0xff;
  kratkaColorB = tkratkaKolor & 0xff;
  kratkaColorA = 0xff;
  playercolorR[0] = 0;
  playercolorG[0] = 0;
  playercolorB[0] = 0;
  playercolorA[0] = 0xff;
  playercolorR[1] = (tplayer1color >> 16) & 0xff;
  playercolorG[1] = (tplayer1color >> 8) & 0xff;
  playercolorB[1] = tplayer1color & 0xff;
  playercolorA[1] = 0xff;
  playercolorR[2] = (tplayer2color >> 16) & 0xff;
  playercolorG[2] = (tplayer2color >> 8) & 0xff;
  playercolorB[2] = tplayer2color & 0xff;
  playercolorA[2] = 0xff;
  kartkaBackgroundColorR = (tkartkaBackgroundColor >> 16) & 0xff;
  kartkaBackgroundColorG = (tkartkaBackgroundColor >> 8) & 0xff;
  kartkaBackgroundColorB = tkartkaBackgroundColor & 0xff;
  kartkaBackgroundColorA = 0xff;
  kropkaRadius = tkropkaRadius;
  if (kropkaRadius < 2) kropkaRadius = 2;
  kropkaActiveBorder = kropkaRadius / 4;
  if (kropkaActiveBorder < 2) kropkaActiveBorder = 2;
  wallThickness = twallThickness;
  if (wallThickness < 1) wallThickness = 1;
  /* get out of here */
  return(0);
}


void ui_drawKartka(int width, int height, int *scrollvector) {
  int x, y;
  gra_clear();  /* clears the whole screen to start fresh */
  /* draw the background color on the entire window */
  gra_getwinsize(&x, &y);
  gra_drawrect(0, 0, x, y, kartkaBackgroundColorR, kartkaBackgroundColorG, kartkaBackgroundColorB, kartkaBackgroundColorA, 1);
  /* draw the horizontal lines */
  for (y = -1; y <= height; y++) {
    if ((y == -1) || (y == height)) {
        gra_drawrect(scrollvector[0], (y + 1) * kratkaSize + scrollvector[1], kratkaThickness + (plansza_width + 1) * kratkaSize, kratkaThickness, 0, 0, 0, 0xff, 1);
      } else {
        gra_drawrect(scrollvector[0], (y + 1) * kratkaSize + scrollvector[1], kratkaThickness + (plansza_width + 1) * kratkaSize, kratkaThickness, kratkaColorR, kratkaColorG, kratkaColorB, kratkaColorA, 1);
    }
  }
  /* draw the vertical lines */
  for (x = -1; x <= width; x++) {
    if ((x == -1) || (x == width)) {
        gra_drawrect((x + 1) * kratkaSize + scrollvector[0], scrollvector[1] + kratkaThickness, kratkaThickness, (plansza_height + 1) * kratkaSize - kratkaThickness, 0, 0, 0, 0xff, 1);
      } else {
        gra_drawrect((x + 1) * kratkaSize + scrollvector[0], scrollvector[1] + kratkaThickness, kratkaThickness, (plansza_height + 1) * kratkaSize - kratkaThickness, kratkaColorR, kratkaColorG, kratkaColorB, kratkaColorA, 1);
    }
  }
}


void ui_applyfaderect(int size) {
  int w, h, step;
  gra_getwinsize(&w, &h);
  for (step = 0; step < size; step += 1) {
    int a = 0xff - ((0xff * step) / size);
    gra_drawrect(step, step, w - (step * 2), h - (step * 2), 0xff, 0xff, 0xff, a, 0);
  }
}


void ui_drawscores(int *scores, char *msg) {
  char scoresstr[128];
  static char *playernames[] = {"YOU", "OPPONENT"};
  int x, w, h;
  struct gra_sprite *sc[2], *msgspr = NULL;
  gra_getwinsize(&w, &h);
  /* load textures */
  for (x = 0; x < 2; x++) {
    snprintf(scoresstr, 256, "%s: %d", playernames[x], scores[x]);
    sc[x] = gra_text2sprite(scoresstr, font_ttf, font_ttf_len, 20, playercolorR[x+1], playercolorG[x+1], playercolorB[x+1], playercolorA[x+1]);
  }
  if ((msg != NULL) && (msg[0] != 0)) msgspr = gra_text2sprite(msg, font_ttf, font_ttf_len, 20, playercolorR[0], playercolorG[0], playercolorB[0], playercolorA[0]);
  /* background */
  gra_drawrect(0, 0, w, gra_getspriteheight(sc[0]), 0xff, 0xff, 0xff, 0xff, 1);
  /* draw the message, if any */
  gra_drawsprite(msgspr, (w - gra_getspritewidth(msgspr)) / 2, 1);
  /* draw the score for Players 1 & 2 */
  for (x = 0; x < 2; x++) {
    if (x == 0) {
        gra_drawsprite(sc[x], 10, 1);
      } else {
        gra_drawsprite(sc[x], w - (gra_getspritewidth(sc[x]) + 10), 1);
    }
  }
  /* free textures */
  gra_freesprite(sc[0]);
  gra_freesprite(sc[1]);
  gra_freesprite(msgspr);
}


void graph_drawKropka(int x, int y, int player, int activeflag, int *scrollvector) {
  if ((player > 0) && (player < 3)) {
    if (activeflag != 0) {
      gra_circle_filled((x+1) * kratkaSize + scrollvector[0], (y+1) * kratkaSize + scrollvector[1], kropkaRadius, 10, 10, 10, 0xff);
      gra_circle_filled((x+1) * kratkaSize + scrollvector[0], (y+1) * kratkaSize + scrollvector[1], kropkaRadius - kropkaActiveBorder, playercolorR[player], playercolorG[player], playercolorB[player], playercolorA[player]);
    } else {
      gra_circle_filled((x+1) * kratkaSize + scrollvector[0], (y+1) * kratkaSize + scrollvector[1], kropkaRadius, playercolorR[player], playercolorG[player], playercolorB[player], playercolorA[player]);
    }
  }
}


void graph_drawWall(int x1, int y1, int x2, int y2, int player, int *scrollvector) {
  if ((player > 0) && (player < 3)) {
    gra_drawline((x1+1) * kratkaSize + scrollvector[0], (y1+1) * kratkaSize + scrollvector[1], (x2+1) * kratkaSize + scrollvector[0], (y2+1) * kratkaSize + scrollvector[1], wallThickness, playercolorR[player], playercolorG[player], playercolorB[player], playercolorA[player]);
  }
}


void graph_mousemap(int mx, int my, int *dotx, int *doty, int *scrollvector) {
  float x, y, x1, y1;
  int intx, inty;
  /* apply the scroll vector to x/y coordinates */
  mx -= scrollvector[0];
  my -= scrollvector[1];
  /* compute the field coordinates */
  x = mx;
  x -= kratkaSize;
  x /= kratkaSize;
  x1 = x + 0.5;  /* a very primitive way of rounding numbers - but works well, and it's fast */
  intx = x1;
  *dotx = intx;
  y = my;
  y -= kratkaSize;
  y /= kratkaSize;
  y1 = y + 0.5;  /* a very primitive way of rounding numbers - but works well, and it's fast */
  inty = y1;
  *doty = inty;
  /* perform some boundaries check, to not take into account clicks into empty spaces */
  if ((intx - x < -0.3) || (intx - x > 0.3) || (inty - y < -0.3) || (inty - y > 0.3)) {
    *dotx = -1;
    *doty = -1;
  }
}


enum ui_events_t ui_getevent(int *x, int *y, int *z, int nofilter) {
  SDL_Event event;
  static int clickx = -1, clicky = -1, motion = 0;
  for (;;) {
    if (SDL_WaitEventTimeout(&event, 250) == 0) return(Event_NONE);
    if (event.type != SDL_MOUSEMOTION) break;
    if ((event.type == SDL_MOUSEMOTION) && (clickx >= 0)) break;
    if (nofilter != 0) break;
  }
  switch (event.type) {
    case SDL_QUIT:
      return(Event_QUIT);
      break;
    case SDL_MOUSEBUTTONDOWN:
      /* when BUTTON DOWN -> remember pos, to differentiate between CLICK and SWEEP later */
      if ((event.button.button == SDL_BUTTON_LEFT) && (event.button.button == 1)) {
        clickx = event.button.x;
        clicky = event.button.y;
      }
      break;
    case SDL_MOUSEMOTION:
      /* check if we are dealing with sweep */
      if ((event.button.button == 1) && ((abs(clickx - event.button.x) > 3) || (abs(clicky - event.button.y) > 3))) {
        motion = 1;
        *x = clickx - event.button.x;
        *y = clicky - event.button.y;
        *z = 1;
        clickx = event.button.x;
        clicky = event.button.y;
        return(Event_SWEEP);
      } else {
        *x = event.button.x;
        *y = event.button.y;
        return(Event_CURSORMOVE);
      }
      break;
    case SDL_MOUSEWHEEL:
      *x = event.wheel.x;
      *y = event.wheel.y;
      return(Event_SCROLL);
      break;
    case SDL_MOUSEBUTTONUP:
      clickx = -1;
      /* if it was a motion, forget it */
      if (motion != 0) {
        motion = 0;
        return(Event_NONE);
      }
      /* otherwise it's a simple click */
      if (event.button.button == SDL_BUTTON_LEFT) {
        *x = event.button.x;
        *y = event.button.y;
        *z = 1;
        return(Event_CLICK);
      }
      if (event.button.button == SDL_BUTTON_RIGHT) {
        *x = event.button.x;
        *y = event.button.y;
        *z = 2;
        return(Event_CLICK);
      }
  }
  return(Event_UNKNOWN);
}


long ui_getticks(void) {
  return(SDL_GetTicks());
}


static int getmenupos(int i, int itemcount, int offset) {
  int w, h;
  gra_getwinsize(&w, &h);
  return(((h - offset) / (itemcount + 1)) * (i + 1) + offset);
}


static int getmenuchoice(int y, int itemcount, int offset, int choiceheight) {
  int pixelpos, i;
  for (i = 0; i < itemcount; i++) {
    pixelpos = getmenupos(i, itemcount, offset);
    if ((y > pixelpos) && (y < pixelpos + choiceheight)) return(i);
  }
  return(-1);
}


enum ui_gamemode_t ui_selectgamemode(void) {
  #define MENUPOSCOUNTS 3
  char *menumsg[] = {"Play against the computer", "Internet play", "Exit"};
  int i, ev, x, y = 0, z, w, h, selected = 0;
  enum ui_gamemode_t res;
  struct gra_sprite *title, *copyright, *menusprite[MENUPOSCOUNTS][2] = {{NULL}}, *frontpic;
  /* allocate textures */
  title = gra_text2sprite("openKropki v" PVER, font_ttf, font_ttf_len, 40, 0, 0, 0, 0xff);
  copyright = gra_text2sprite("Copyright © Mateusz Viste 2015", font_ttf, font_ttf_len, 20, 0, 0, 0, 0x90);
  frontpic = gra_loadgzbmp(frontpic_bmp_gz, frontpic_bmp_gz_len);
  for (i = 0; i < MENUPOSCOUNTS; i++) {
    menusprite[i][0] = gra_text2sprite(menumsg[i], font_ttf, font_ttf_len, 30, 0, 0, 0, 0xff);
    menusprite[i][1] = gra_text2sprite(menumsg[i], font_ttf, font_ttf_len, 30, 0, 0, 0xff, 0xff);
  }
  /* wait for events and draw screen */
  while (selected == 0) {
    gra_clear();
    gra_getwinsize(&w, &h);
    gra_drawrect(0, 0, w, h, 0, 0, 0, 0xff, 1);
    gra_drawrect(0, 0, w, h, 0xff, 0xff, 0xff, 0xff, 1);
    gra_drawsprite(frontpic, 0, h - gra_getspriteheight(frontpic));
    z = 10;
    gra_drawsprite(title, (w - gra_getspritewidth(title)) >> 1, z);
    z += gra_getspriteheight(title);
    gra_drawsprite(copyright, (w - gra_getspritewidth(copyright)) >> 1, z);
    /* display menu entries */
    for (i = 0; i < MENUPOSCOUNTS; i++) {
      int itemcolor = 0;
      z = getmenupos(i, MENUPOSCOUNTS, gra_getspriteheight(title));
      if (getmenuchoice(y, MENUPOSCOUNTS, gra_getspriteheight(title), gra_getspriteheight(menusprite[0][0])) == i) itemcolor = 1;
      gra_drawsprite(menusprite[i][itemcolor], (w - gra_getspritewidth(menusprite[i][0])) >> 1, z);
    }
    /* refresh screen and look for events */
    gra_refresh();
    ev = ui_getevent(&x, &y, &z, 1);
    switch (ev) {
      case Event_CLICK:
        i = getmenuchoice(y, MENUPOSCOUNTS, gra_getspriteheight(title), gra_getspriteheight(menusprite[0][0]));
        if (i == 0) {
          res = GAMEMODE_AI;
          selected = 1;
        } else if (i == 1) {
          res = GAMEMODE_NET;
          selected = 1;
        } else if (i == 2) {
          res = GAMEMODE_QUIT;
          selected = 1;
        }
        break;
      case Event_QUIT:
        res = GAMEMODE_QUIT;
        selected = 1;
        break;
    }
  }
  /* cleanup textures and exit */
  gra_freesprite(title);
  gra_freesprite(copyright);
  gra_freesprite(frontpic);
  for (i = 0; i < MENUPOSCOUNTS; i++) {
    gra_freesprite(menusprite[i][0]);
    gra_freesprite(menusprite[i][1]);
  }
  return(res);
  #undef MENUPOSCOUNTS
}


void ui_throwmsg(char *msg) {
  int w, h, x, y;
  struct gra_sprite *msgspr;
  gra_getwinsize(&w, &h);
  msgspr = gra_text2sprite(msg, font_ttf, font_ttf_len, 20, 0, 0, 0, 0xff);
  x = (w - gra_getspritewidth(msgspr)) / 2;
  y = (h - gra_getspriteheight(msgspr)) / 2;
  gra_drawsprite(msgspr, x, y);
  gra_refresh();
  gra_freesprite(msgspr);
}


void ui_delay(int ms) {
  SDL_Delay(ms);
}


void ui_changecursor(enum ui_mousecursor_t curid) {
  static SDL_Cursor *oldcurs, *curs;
  switch (curid) {
    case CURSOR_WAIT:
      curs = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_WAIT);
      break;
    default:
      curs = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_ARROW);
      break;
  }
  SDL_SetCursor(curs);
  if (oldcurs != NULL) SDL_FreeCursor(oldcurs);
  oldcurs = curs;
}
